/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, SAMSUNG ELECTRONICS CO., LTD. and BRITISH BROADCASTING CORPORATION
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of SAMSUNG ELECTRONICS CO., LTD. nor the name of the BRITISH BROADCASTING CORPORATION
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TComPrediction.h
    \brief    prediction class (header)
*/

#ifndef __TCOMPREDICTION__
#define __TCOMPREDICTION__


// Include files
#include "TComPic.h"
#include "TComMotionInfo.h"
#include "TComPattern.h"
#include "TComTrQuant.h"
#include "TComPredFilter.h"

#ifdef EDGE_BASED_PREDICTION
#include "TComEdgeBased.h"
#endif //EDGE_BASED_PREDICTION

#if HHI_INTERP_FILTER
#include "TComPredFilterMOMS.h"
#endif

#if (MC_MEMORY_ACCESS_CALC && MC_MEMORY_ACCESS_CACHE)
// Cache line size in bytes - 64 bytes
#define LOG2_CACHE_LINE_SIZE (6)
#define CACHE_LINE_SIZE (1<<LOG2_CACHE_LINE_SIZE)
// Number of lines in cache
#define CACHE_NUM_LINES (2048)
// Total cache size in bytes
// #define CACHE_SIZE (CACHE_LINE_SIZE*CACHE_NUM_LINES)

// Cache types
#define  CACHE_DIRECT_MAPPED              0
#define  CACHE_TWO_WAY                    1
#define  CACHE_FOUR_WAY                   2
#define  CACHE_EIGHT_WAY                  3
#define  CACHE_FULL_ASSOCIATIVITY         4

// Structure definition for cache entry
typedef struct CacheElement {
  Int iValid;            // Indicates whether cache element is valid
  UInt64 ui64LineAddress;    // Address of line being stored
  Int iLRUCount;
} CACHE_ELEMENT;
#endif

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// prediction class
class TComPrediction : public TComPredFilter
#if HHI_INTERP_FILTER
                      ,public TComPredFilterMOMS
#endif
{
protected:
  Int*      m_piYuvExt;
  Int       m_iYuvExtStride;
  Int       m_iYuvExtHeight;
  Int       m_iDIFHalfTap;
#ifdef EDGE_BASED_PREDICTION
  Int*      m_piYExtEdgeBased;
#endif //EDGE_BASED_PREDICTION
#if MC_MEMORY_ACCESS_CALC
  Double    m_dMemCompressionRatio;
  Int       m_iMemCompressionUnitWidth;
  Int       m_iMemCompressionUnitHeight;
  Double    m_dChromaMemCompressionRatio;
  Int       m_iChromaMemCompressionUnitWidth;
  Int       m_iChromaMemCompressionUnitHeight;
#if MC_MEMORY_ACCESS_CACHE
  /*
   * Define cache of size CACHE_NUM_LINES. First dimension of 5 is for 
   * defining separated cache for different block sizes (1x1, 1x8, 4x4, 
   * 8x8, and custom)
   */
  CACHE_ELEMENT Cache[5][CACHE_NUM_LINES];

/*
 * piCachePtr: Pointer that points to cache location to fill next.
 * One pointer per partition (number of partitions defined by
 * associativity). For full associativity, we use only one pointer.
 * For direct mapped we use CACHE_NUM_LINES number of pointers. For 2-way
 * cache we use CACHE_NUM_LINES/2 pointers.
 */
  Int piCachePtr[5][CACHE_NUM_LINES];
#endif //  MC_MEMORY_ACCESS_CACHE
#endif // MC_MEMORY_ACCESS_CALC
  TComYuv   m_acYuvPred[2];
  TComYuv   m_cYuvPredTemp;
  TComYuv   m_cYuvExt;

  // ADI functions
  Void xPredIntraAngleAdi       ( Int* pSrc, Int iSrcStride, Pel*& tpDst, Int iDstStride, UInt iWidth, UInt iHeight, UInt uiDirMode );
  Void xPredIntraVertAdi        ( Int* pSrc, Int iSrcStride, Pel*& tpDst, Int iDstStride, UInt iWidth, UInt iHeight );
  Void xPredIntraHorAdi         ( Int* pSrc, Int iSrcStride, Pel*& tpDst, Int iDstStride, UInt iWidth, UInt iHeight );
  Void xPredIntraDCAdi          ( Int* pSrc, Int iSrcStride, Pel*& tpDst, Int iDstStride, UInt iWidth, UInt iHeight, Bool bAboveAvail, Bool bLeftAvail );
  Void xPredIntraPlaneAdi       ( Int* pSrc, Int iSrcStride, Pel*& tpDst, Int iDstStride, UInt iWidth, UInt iHeight );
  Void xPredIntraBiAdi          ( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, UInt iWidth, UInt iHeight );
  __inline Int xGetContextPixel ( UInt uiDirMode, UInt uiXYflag, Int iTempD, Bool& bCenter );

#if PLANAR_INTRA
  Void xPredIntraPlanar         ( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, UInt iWidth, UInt iHeight, Int iSampleBottomRight, Bool bAbove, Bool bLeft );
#endif

#if ANG_INTRA
  Void xPredIntraAng            ( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, UInt iWidth, UInt iHeight, UInt uiDirMode, Bool bAbove, Bool bLeft );
#endif

  // motion compensation functions
  Void xPredInterUni            ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Int iPartIdx          );
  Void xPredInterBi             ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight,                         TComYuv*& rpcYuvPred, Int iPartIdx          );
  Void xPredInterLumaBlk        ( TComDataCU* pcCU, TComPicYuv* pcPicYuvRef, UInt uiPartAddr, TComMv* pcMv, Int iWidth, Int iHeight,                         TComYuv*& rpcYuv );
  Void xPredInterChromaBlk      ( TComDataCU* pcCU, TComPicYuv* pcPicYuvRef, UInt uiPartAddr, TComMv* pcMv, Int iWidth, Int iHeight,                         TComYuv*& rpcYuv                            );
#if TEN_DIRECTIONAL_INTERP
  Void xPredInterLumaBlk_TEN    ( TComDataCU* pcCU, TComPicYuv* pcPicYuvRef, UInt uiPartAddr, TComMv* pcMv, Int iWidth, Int iHeight,                         TComYuv*& rpcYuv );
  Void xPredInterChromaBlk_TEN  ( TComDataCU* pcCU, TComPicYuv* pcPicYuvRef, UInt uiPartAddr, TComMv* pcMv, Int iWidth, Int iHeight,                         TComYuv*& rpcYuv                            );
#endif
  Void xWeightedAverage         ( TComDataCU* pcCU, TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst );
#ifdef EDGE_BASED_PREDICTION
  TComEdgeBased* EdgeBasedPred;  
#endif //EDGE_BASED_PREDICTION
#ifdef QC_AMVRES
  Void  xPredInterLumaBlkHMV ( TComDataCU* pcCU, TComPicYuv* pcPicYuvRef, UInt uiPartAddr, TComMv* pcMv,Int iWidth, Int iHeight, TComYuv*& rpcYuv );
  Void  xPredInterLumaBlkHMVME ( Pel* piSrcY, Int iSrcStride, Pel* piDstY, Int iDstStride, TComMv* pcMv,Int iWidth, Int iHeight, Int dMVx, Int dMVy);
  Void  xHAM_Filter (Pel*  piRefY, Int iRefStride,Pel*  piDstY,Int iDstStride, Int iWidth, Int iHeight,Int iMVyFrac,Int iMVxFrac);
#endif
#ifdef QC_SIFO
  Void xSIFOFilter              (Pel*  piRefY, Int iRefStride,Pel*  piDstY,Int iDstStride, Int iWidth, Int iHeight,Int iyFrac,Int ixFrac);
  Void xPredInterLumaBlk_SIFOApplyME ( Pel* piSrcY, Int iSrcStride, Pel* piDstY, Int iDstStride, TComMv* pcMv, 
                                                   Int iWidth, Int iHeight, Pel* piOrg,Int iStrideOrg, Int dMVx, Int dMVy);
#endif
#if MC_MEMORY_ACCESS_CALC
  Int iAlignmentRoundDown        (Int iAddr, Int iAccessUnit);
  Int iAlignmentRoundUp          (Int iAddr, Int iAccessUnit);
  Int iUncompressedMemCalc1D     (Int iMCBlockWidth, 
                                  Int iMCBlockHeight, 
                                  Int iMCBlockPosX,
                                  Int iMCBlockPosY,
                                  Int iMCHorizontalIFTap, 
                                  Int iMCVerticalIFTap, 
                                  Int iSampleBits,
                                  Int iDDRAlignBits,
							      Int iDDRBurstBits,
								  Int iPicWidth,
                                  Bool bMC_uv_flag);
  Int iCompressedMemCalc1DStrage (Int iMCBlockWidth, 
                                  Int iMCBlockHeight, 
                                  Int iMCBlockPosX,
                                  Int iMCBlockPosY,
                                  Int iMCHorizontalIFTap, 
                                  Int iMCVerticalIFTap, 
                                  Int iCompressionUnitWidth,
                                  Int iCompressionUnitHeight,
                                  Int iSampleBits,
                                  Int iDDRAlignBits,
                                  Int iDDRBurstBits,
                                  Int iPicWidth,
                                  Double dCompressionRatio,
                                  Bool bMC_uv_flag);
#if MC_MEMORY_ACCESS_CACHE
  Void xCacheClear();
  Void xCalcCachePartition(UInt64 ui64Address, Int associativity, Int *partition, Int *subCacheSize);
  Int iCalcCacheHitMissBW(Int associativity, UInt64 ui64Address, Int iCacheNum);
  Void xCacheInsert(Int iCacheNum, Int partition, Int subCacheSize, UInt64 ui64LineAddress);
  Void xCacheUpdate(Int iCacheNum, Int partition, Int subCacheSize, Int iLocation);
  Int iCompressedMemCalc1DStrageCache (Int iMCBlockWidth, 
                                  Int iMCBlockHeight, 
                                  Int iMCBlockPosX,
                                  Int iMCBlockPosY,
                                  Int iMCHorizontalIFTap, 
                                  Int iMCVerticalIFTap, 
                                  Int iCompressionUnitWidth,
                                  Int iCompressionUnitHeight,
                                  Int iSampleBits,
                                  Int iDDRAlignBits,
                                  Int iDDRBurstBits,
                                  Int iPicWidth,
                                  Double dCompressionRatio,
                                  Bool bMC_uv_flag,
								  Int iPoc,
								  Int iCacheNum);
  Int iUncompressedMemCalc1DCache(Int iMCBlockWidth, 
                                  Int iMCBlockHeight, 
                                  Int iMCBlockPosX,
                                  Int iMCBlockPosY,
                                  Int iMCHorizontalIFTap, 
                                  Int iMCVerticalIFTap, 
                                  Int iSampleBits,
                                  Int iDDRAlignBits,
								  Int iDDRBurstBits,
								  Int iPicWidth,
                                  Bool bMC_uv_flag,
								  Int iPoc,
								  Int iCacheNum);
#endif
  Void  xMemCalc                 (TComDataCU* pcCU,
	                              Int iMCBlockWidth, 
                                  Int iMCBlockHeight, 
                                  Int iMCBlockPosX,
                                  Int iMCBlockPosY,
                                  Int iMCHorizontalIFTap, 
                                  Int iMCVerticalIFTap, 
                                  Int iCompressionUnitWidth,
                                  Int iCompressionUnitHeight,
								  Double dCompressionRatio,
                                  Int iSampleBits,
                                  Int iPicWidth,
                                  Bool bMC_uv_flag);
#if MC_MEMORY_ACCESS_CACHE                                  
  Void  xMemCalcCache             (TComDataCU* pcCU,
	                              Int iMCBlockWidth, 
                                  Int iMCBlockHeight, 
                                  Int iMCBlockPosX,
                                  Int iMCBlockPosY,
                                  Int iMCHorizontalIFTap, 
                                  Int iMCVerticalIFTap, 
                                  Int iCompressionUnitWidth,
                                  Int iCompressionUnitHeight,
								  Double dCompressionRatio,
                                  Int iSampleBits,
                                  Int iPicWidth,
                                  Bool bMC_uv_flag,
								  Int iPoc);
#endif                                  
  Void xMemAccessHighCodingEfficiencyAnchor (TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight);
  Void xMemAccessLowComplexityAnchor        (TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight);
#endif MC_MEMORY_ACCESS_CALC
public:
  TComPrediction();
  virtual ~TComPrediction();

  Void    initTempBuff();

  // inter
  Void motionCompensation         ( TComDataCU*  pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList = REF_PIC_LIST_X, Int iPartIdx = -1 );

  // motion vector prediction
  Void getMvPredAMVP              ( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv& rcMvPred );

#ifdef HHI_IMVP
  Void getMvPredIMVP( TComDataCU* pcSubCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefList, Int iRefIdx, TComCUMvField* pcSubCUMvField, TComMv& rcMvPred );
#ifdef QC_AMVRES
  Void getMvPredIMVP_onefourth( TComDataCU* pcSubCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefList, Int iRefIdx, TComCUMvField* pcSubCUMvField, TComMv& rcMvPred );
#endif
#endif

  // ADI
#if HHI_AIS
  Void predIntraLumaAdi           ( TComPattern* pcTComPattern,             UInt uiDirMode, Bool bSmoothing, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight,  TComDataCU* pcCU, Bool bAbove, Bool bLeft );
#else
  Void predIntraLumaAdi           ( TComPattern* pcTComPattern,             UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight,  TComDataCU* pcCU, Bool bAbove, Bool bLeft );
#endif
  Void predIntraChromaAdi         ( TComPattern* pcTComPattern, Int* piSrc, UInt uiDirMode,                  Pel* piPred, UInt uiStride, Int iWidth, Int iHeight,  TComDataCU* pcCU, Bool bAbove, Bool bLeft );

  // CIP
  Void recIntraLumaCIP            ( TComPattern* pcTComPattern, Pel* pPred, Pel* pResi, Pel* pReco, UInt uiStride, Int iWidth, Int iHeight, TComDataCU* pcCU, Bool bAboveAvail, Bool bLeftAvail );

#if ANG_INTRA
  // Angular Intra
#if HHI_AIS
  Void predIntraLumaAng           ( TComPattern* pcTComPattern, UInt uiDirMode, Bool bSmoothing, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight,  TComDataCU* pcCU, Bool bAbove, Bool bLeft );
#else
  Void predIntraLumaAng           ( TComPattern* pcTComPattern, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight,  TComDataCU* pcCU, Bool bAbove, Bool bLeft );
#endif
  Void predIntraChromaAng         ( TComPattern* pcTComPattern, Int* piSrc, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, TComDataCU* pcCU, Bool bAbove, Bool bLeft );
#endif

#if (ANG_INTRA || PLANAR_INTRA)
  Pel  predIntraGetPredValDC      ( Int* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, Bool bAbove, Bool bLeft );
#endif

#if PLANAR_INTRA
  Void predIntraPlanar            ( Int* piSrc, Int iDelta, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft );
#endif

  Int* getPredicBuf()             { return m_piYuvExt;      }
  Int  getPredicBufWidth()        { return m_iYuvExtStride; }
  Int  getPredicBufHeight()       { return m_iYuvExtHeight; }
#ifdef EDGE_BASED_PREDICTION
  TComEdgeBased* getEdgeBasedPred () { return EdgeBasedPred; }
  Int* getEdgeBasedBuf()          { return m_piYExtEdgeBased; }
#endif //EDGE_BASED_PREDICTION

#if MC_MEMORY_ACCESS_CALC
  Void setMemCompressionRatio(Double dCompressionRatio)                  { m_dMemCompressionRatio = dCompressionRatio;};
  Void setMemCompressionUnitWidth (Int iMemCompressionUnitWidth)         { m_iMemCompressionUnitWidth = iMemCompressionUnitWidth;};
  Void setMemCompressionUnitHeight(Int iMemCompressionUnitHeight)        { m_iMemCompressionUnitHeight = iMemCompressionUnitHeight;};
  Void setChromaMemCompressionRatio(Double dChromaCompressionRatio)                  { m_dChromaMemCompressionRatio = dChromaCompressionRatio;};
  Void setChromaMemCompressionUnitWidth (Int iChromaMemCompressionUnitWidth)         { m_iChromaMemCompressionUnitWidth = iChromaMemCompressionUnitWidth;};
  Void setChromaMemCompressionUnitHeight(Int iChromaMemCompressionUnitHeight)        { m_iChromaMemCompressionUnitHeight = iChromaMemCompressionUnitHeight;};
#endif // MC_MEMORY_ACCESS_CALC
};


#endif // __TCOMPREDICTION__

